{{
        Test RealRandomSpin

        Porting Chip Gracey's code from PASM to Spin

        Jonathan Dummer
}}
                       
CON                    
  ' Clock Mode
  _clkmode        = xtal1 + pll16x
  _xinfreq        = 5_000_000

  ' Inform Lord Helmet!
  COM_BAUD = 460_800 ' 460_800 ' 9_600, 57_600, 115_200, 230_400, 250_000, 460_800

      
OBJ
  term:         "FFDS1"
  rr:           "RealRandom"

VAR
  long  histo[256]
  
PUB TestRRS : tot | i, dt, mx

  ' serial link back to the PC 
  term.Start( 31, 30, COM_BAUD ) 
         
  repeat

    ' run a bunch
    dt := -cnt
    repeat 1000
      histo[ RealRandomSpin( 3 ) ] += 1
    dt += cnt
    tot += 1000

    term.str( string( 13, 13, "RRS Histogram: [ms] = " ) )
    dt /= clkfreq / 1000
    term.dec( dt )
    mx := histo
    repeat i from 1 to 255
      mx #>= histo[i]    
    repeat i from 0 to 255
      if histo[i]
        term.tx( 13 )
        if i < 100
          term.tx( " " )
        if i < 10
          term.tx( " " )
        term.dec( i )
        term.tx( ":" )
        dt := (histo[i] * 32 + (mx >> 1)) / mx
        repeat dt
          term.tx( "=" )
        repeat 33-dt
          term.tx( " " )
        term.dec( histo[ i ] )
        
        
           
PUB RealRandomSpin( bits ) : rn | rmask, rbit_old, rbit_new, got_subbits

  ' set up the counter
  ctra := constant( %00001_111 << 23 )                  'set ctra to internal pll mode, select x16 tap    
  frqa := constant( $020 << 23 )                        'set frqa to system clock frequency / 16 
  vcfg := constant( $040 << 23 )                        'set vcfg to discrete output, but without pins
  vscl := 70                                            'set vscl to 70 pixel clocks per waitvid

  repeat bits

    ' flag that we don't have any sub-bits yet
    got_subbits := 0
    
    ' get one random bit
    repeat

      ' get at least 2 sub-bits, and make sure they don't match!   
      repeat
        rbit_old := rbit_new
        waitvid( 0, 0 )
        rmask := phsa & %10111
        rbit_new := rmask
        ' basically checking if the number of 1's in rmask is odd or even
        repeat 
          rbit_new ^= (rmask >>= 1)        
        while rmask
        rbit_new &= 1
        ' work that odd or even back into phsa (rcr), and then mix in the counter
        phsa := (phsa & -2 + rbit_new) -> 1 + cnt
      until got_subbits++
      
    until rbit_old ^ rbit_new
    
    ' save this latest bit
    rn += rn + rbit_new
  

{{
  Reference code from RealRandom by Chip Gracey
}}
{
DAT
                        org

entry                   movi    ctra,#%00001_111        'set ctra to internal pll mode, select x16 tap
                        movi    frqa,#$020              'set frqa to system clock frequency / 16
                        movi    vcfg,#$040              'set vcfg to discrete output, but without pins
                        mov     vscl,#70                'set vscl to 70 pixel clocks per waitvid

:twobits                waitvid 0,0                     'wait for next 70-pixel mark  jitter time
                        test    phsa,#%10111    wc      'pseudo-randomly sequence phase to induce jitter
                        rcr     phsa,#1                 '(c holds random bit #1)
                        add     phsa,cnt                'mix PLL jitter back into phase 

                        rcl     par,#1          wz, nr  'transfer c into nz (par shadow register = 0)
                        wrlong  _random_value,par       'write random value back to spin variable

                        waitvid 0,0                     'wait for next 70-pixel mark  jitter time           
                        test    phsa,#%10111    wc      'pseudo-randomly sequence phase to induce jitter        
                        rcr     phsa,#1                 '(c holds random bit #2)                                                        
                        add     phsa,cnt                'mix PLL jitter back into phase                    

        if_z_eq_c       rcl     _random_value,#1        'only allow different bits (removes bias)
                        jmp     #:twobits               'get next two bits


_random_value           res     1
}             